/**
 * Copyright (c) 2014-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

@link-color: #4183C4;
@header-color: #212325;
@body-color: #626466;

html, body {
  -webkit-tap-highlight-color: rgba(0, 0, 0, 0);
  -webkit-text-size-adjust: 100%;
  -ms-text-size-adjust: 100%;
  margin: 0;
  padding: 0;
  -webkit-font-smoothing: antialiased;
  color: @body-color;
  font-family: 'Fira Sans', 'Helvetica Neue', Helvetica, Arial, sans-serif;
  font-size: 16px;
  line-height: 1.625;
}

@media only screen and (max-width: 680px) {
  body {
    font-size: 14px;
  }
}

h1, h2, h3, h4, h5, h6 {
  color: @header-color;
}

a {
  color: @link-color;
  text-decoration: none;
}

pre, code {
  font-family: 'Fira Mono', Menlo, monospace;
  background: #F9F8F7;
  color: #484A4C;
  font-size: 0.9375em;
  letter-spacing: -0.015em;
}

code {
  margin: -0.05rem -0.15em;
  padding: 0.05rem 0.35em;
}

blockquote {
  margin: 1rem 0;
  padding: 0 1rem;
  color: #727476;
  border-left: solid 3px #DCDAD9;
}

blockquote > :first-child {
  margin-top: 0;
}

blockquote > :last-child {
  margin-bottom: 0;
}


// Markdown

.codeBlock {
  -webkit-overflow-scrolling: touch;
  background: #FCFBFA;
  border-left: solid 3px #ECEAE9;
  box-sizing: border-box;
  display: block;
  font-size: 0.875em;
  margin: 0.5rem 0;
  overflow-y: scroll;
  padding: 0.5rem 8px 0.5rem 12px;
  white-space: pre;
}

.t.blockParams {
  padding-left: 2ch;
}

// TODO: not random colors

.token.punctuation,
.token.ignore,
.t.interfaceDef,
.t.member,
.t.callSig {
  color: #808890;
}

.token.function,
.token.class-name,
.token.qualifier,
.t.fnQualifier,
.t.fnName {
  color: #32308E;
}

.token.primitive,
.t.primitive {
  color: #922;
}

.token.number,
.t.typeParam {
  color: #905;
}

.t.typeQualifier,
.t.typeName {
  color: #013679;
}

.t.param {
  color: #945277;
}

.t.memberName {
  color: teal;
}

.token.block-keyword,
.token.keyword,
.t.keyword {
  color: #A51;
}

.token.string,
.token.regex {
  color: #df5050;
}

.token.operator {
  color: #a67f59;
}

.token.comment {
  color: #998;
  font-style: italic;
}
